#include <math.h>
#include "st7920.h"

#define LCD_WIDTH 128
#define LCD_HEIGHT 64

#define NO_GAMES 2
#define SNAKE 0
#define PONG 1

#define MAX_SNAKE_LENGTH 512
#define INIT_SNAKE_LENGTH 10
#define SNAKE_LCD_WIDTH 64
#define SNAKE_LCD_HEIGHT 32

#define PONG_PADDLE_LENGTH 20
#define PONG_BALL_SIZE 4
#define NO_PONG_GAMES_TO_WIN 5



uint8_t up, down, left, right, action, menu;

char padding_string[17] = "                ";
char you_lost_string[17] = "    You Lost!   ";
char you_won_string[17] = "    You Won!    ";
char game_name[NO_GAMES][17] = {"       Snake    ",
								"       Pong     "};

void init_buttons() {
	DDRA &= ~(1 << PA0);
	PORTA |= (1 << PA0);
	DDRA &= ~(1 << PA1);
	PORTA |= (1 << PA1);
	DDRA &= ~(1 << PA2);
	PORTA |= (1 << PA2);
	DDRA &= ~(1 << PA3);
	PORTA |= (1 << PA3);
	DDRA &= ~(1 << PA4);
	PORTA |= (1 << PA4);
	DDRA &= ~(1 << PA5);
	PORTA |= (1 << PA5);
}

void check_input() {
	menu = 0;
	action = 0;
	up = 0;
	down = 0;
	left = 0;
	right = 0;

	if ((PINA & (1 << PA0)) == 0)
		menu = 1;
	if ((PINA & (1 << PA1)) == 0)
		action = 1;
	if ((PINA & (1 << PA2)) == 0)
		down = 1;
	if ((PINA & (1 << PA3)) == 0)
		up = 1;
	if ((PINA & (1 << PA4)) == 0)
		right = 1;
	if ((PINA & (1 << PA5)) == 0)
		left = 1;
}

//waits untill a button is pressed
uint8_t wait_for_input() {
	uint8_t t = 0;
	while(1) {
		check_input();
		if(menu || action || up || down || left || right)
			break;
		t++;
	}
	return t;
}

void beep(uint8_t type, uint8_t length) {
	int i;
	for(i = 0; i < length * 10; i++) {
		PORTD |= 1 << PD5;
		_delay_ms (1);
		PORTD &= ~(1 << PD5);
        _delay_ms(type);
    }
}

void win_sound() {
	beep(1, 10);
    _delay_ms(10);
    beep(1, 10);
    _delay_ms(10);
    beep(1, 10);
    _delay_ms(10);
}

void lose_sound() {
	beep(4, 15);
}

uint8_t game_menu() {
	uint8_t game = SNAKE;

	clcd_init();
	display_text(padding_string, padding_string, padding_string, padding_string);
	display_text(padding_string, game_name[game], padding_string, padding_string);

	while(1) {
		check_input();

		//start game
		if(action)
			break;

		//next
		if(right) {
			while(right)
				check_input();
			
			game = (game + 1) % NO_GAMES;
			display_text(padding_string, game_name[game], padding_string, padding_string);
		}

		//prev
		if(left) {
			while(left)
				check_input();

			if(game == 0)
				game = NO_GAMES - 1;
			else
				game--;
			display_text(padding_string, game_name[game], padding_string, padding_string);
		}

		//go to the first game
		if(menu && game != SNAKE) {
			while(menu)
				check_input();

			game = SNAKE;
			display_text(padding_string, game_name[game], padding_string, padding_string);
		}

		_delay_ms(100);
	}
	_delay_ms(500);
	cdisplay_clear();

	return game;
}

//check for collisions
uint8_t snake_ok(int i, int j, int tail, int head, uint8_t *x, uint8_t *y) {
	if(i < 0 || j < 0 || i >= SNAKE_LCD_WIDTH || j >= SNAKE_LCD_HEIGHT)
		return 0;

	int k;
	for(k = tail + 1; k < MAX_SNAKE_LENGTH && k != head; k++)
		if(x[k] == i && y[k] == j)
			return 0;

	if(k == MAX_SNAKE_LENGTH)
		for(k = 0; k < head; k++)
			if(x[k] == i && y[k] == j)
				return 0;

	return 1;
}

//the point that you try to eat
void get_random_point(int tail, int head, uint8_t *x, uint8_t *y, uint8_t *point_x, uint8_t *point_y) {
	uint8_t ok, px, py;
	int k;

	do {
		ok = 1;
		px = rand() % SNAKE_LCD_WIDTH;
		py = rand() % SNAKE_LCD_HEIGHT;

		for(k = tail + 1; k < MAX_SNAKE_LENGTH && k != head; k++)
			if(x[k] == px && y[k] == py) {
				ok = 0;
				break;
			}

		if(k == MAX_SNAKE_LENGTH && ok)
			for(k = 0; k < head; k++)
				if(x[k] == px && y[k] == py) {
					ok = 0;
					break;
				}
	}while(!ok);

	*point_x = px;
	*point_y = py;
}

void snake() {
	uint8_t x[MAX_SNAKE_LENGTH], y[MAX_SNAKE_LENGTH];
	uint8_t start_x, start_y, point_x, point_y;
	uint8_t undelay;
	int head, tail, length;
	int i, j, di, dj;

	length = INIT_SNAKE_LENGTH;
	
	tail = 0;
	head = length;

	start_x = 15;
	start_y = 15;

	for(i = 0; i < length; i++) {
		x[i] = start_x + i;
		y[i] = start_y;
	}

	di = 1;
	dj = 0;
   
	i = start_x + length;
	j = start_y;


	glcd_enable();
	glcd_fill(0x0000);

	draw_horiz_line(2 * start_x, 2 * start_y, 2 * length);
	draw_horiz_line(2 * start_x, 2 * start_y + 1, 2 * length);

	srand(wait_for_input());

	if(menu) {
		glcd_fill(0x0000);
		return;
	}

	get_random_point(tail, head, x, y, &point_x, &point_y);
	draw_block(point_x, point_y);

	while(!menu) {
		if(!snake_ok(i, j, tail, head, x, y)) {
			lose_sound();
			
			glcd_fill(0x0000);
			clcd_init();
			display_text(padding_string, you_lost_string, padding_string, padding_string);		    
		    wait_for_input();
		    cdisplay_clear();
		    _delay_ms(200);
			return;
		}


		if(i != point_x || j != point_y || length == MAX_SNAKE_LENGTH) {
			clear_block(x[tail], y[tail]);
			tail = (tail + 1) % MAX_SNAKE_LENGTH;
		} else {
			length++;
			if(length == MAX_SNAKE_LENGTH) {
				glcd_fill(0x0000);
				clcd_init();
				display_text(padding_string, you_won_string, padding_string, padding_string);		    
			    win_sound();
			    wait_for_input();
			    cdisplay_clear();
			    _delay_ms(200);
			    return;
			}
		}

		//eat the point
		undelay = 0;
		if(i == point_x && j == point_y) {
			beep(1, 1);
			get_random_point(tail, head, x, y, &point_x, &point_y);
			draw_block(point_x, point_y);
			undelay = 50;
		}
	   
		draw_block(i, j);
		
		x[head] = i;
		y[head] = j;
		head = (head + 1) % MAX_SNAKE_LENGTH;

		i += di;
		j += dj;

		check_input();

		if (down && (di != 0 ||  dj != -1)) {
			di = 0; dj = 1;
		} else if (up && (di != 0 ||  dj != 1)) {
			di = 0; dj = -1;
		} else if (right && (di != -1 ||  dj != 0)) {
			di = 1; dj = 0;
		} else if (left && (di != 1 ||  dj != 0)) {
			di = -1; dj = 0;
		}
		
		if(action)
			_delay_ms(50 - undelay);
		else
			_delay_ms(100 - undelay);   
	}

	glcd_fill(0x0000);
}

void draw_ball(uint8_t ball_x, uint8_t ball_y) {
	int i, j;

	for(i = ball_x; i < ball_x + PONG_BALL_SIZE; i++)
		for(j = ball_y; j < ball_y + PONG_BALL_SIZE; j++)
			set_pixel(i, j);

}

void move_ball(uint8_t *ball_x, uint8_t *ball_y, int dx, int dy) {
	uint8_t i, j;

	if(dx >= 0)
		for(i = 0; i < dx; i++)
			for(j = 0; j < PONG_BALL_SIZE; j++) 
				clr_pixel(*ball_x + i, *ball_y + j);
	else
		for(i = 0; i < -dx; i++)
			for(j = 0; j < PONG_BALL_SIZE; j++) 
				clr_pixel(*ball_x + PONG_BALL_SIZE - i - 1, *ball_y + j);

	if(dy >= 0)
		for(j = 0; j < dy; j++)
			for(i = 0; i < PONG_BALL_SIZE; i++)
				clr_pixel(*ball_x + i, *ball_y + j);
	else
		for(j = 0; j < -dy; j++)
			for(i = 0; i < PONG_BALL_SIZE; i++)
				clr_pixel(*ball_x + i, *ball_y + PONG_BALL_SIZE - j - 1);

	if(dx >= 0)
		for(i = 0; i < dx; i++)
			for(j = 0; j < PONG_BALL_SIZE; j++) 
				set_pixel(*ball_x + PONG_BALL_SIZE + i, *ball_y + dy + j);
	else
		for(i = 0; i < -dx; i++)
			for(j = 0; j < PONG_BALL_SIZE; j++) 
				set_pixel(*ball_x - i - 1, *ball_y + dy + j);

	if(dy >= 0)
		for(j = 0; j < dy; j++)
			for(i = 0; i < PONG_BALL_SIZE; i++)
				set_pixel(*ball_x + dx + i, *ball_y + PONG_BALL_SIZE + j);
	else
		for(j = 0; j < -dy; j++)
			for(i = 0; i < PONG_BALL_SIZE; i++)
				set_pixel(*ball_x + dx + i, *ball_y - j - 1);

	*ball_x += dx;
	*ball_y += dy;
}

uint8_t move_paddle_up(uint8_t start_x, uint8_t start_y) {
	if(start_y > 0) {
		clr_pixel(start_x, start_y + PONG_PADDLE_LENGTH - 1);
		clr_pixel(start_x + 1, start_y + PONG_PADDLE_LENGTH - 1);
		set_pixel(start_x, start_y - 1);
		set_pixel(start_x + 1, start_y - 1);
		start_y--;
	}
	return start_y;
}

uint8_t move_paddle_down(uint8_t start_x, uint8_t start_y) {
	if(start_y + PONG_PADDLE_LENGTH < LCD_HEIGHT) {
		clr_pixel(start_x, start_y);
		clr_pixel(start_x + 1, start_y);
		set_pixel(start_x, start_y + PONG_PADDLE_LENGTH);
		set_pixel(start_x + 1, start_y + PONG_PADDLE_LENGTH);
		start_y++;
	}
	return start_y;
}

int detect_collision(uint8_t *ball_x, uint8_t *ball_y, int *dx, int *dy, uint8_t j1, uint8_t j2) {
	int x, y, offset;

	x = *ball_x + *dx;
	y = *ball_y + *dy;

	//check player's paddle
	if(x < 2) {
		if(y + PONG_BALL_SIZE <= j1 - 1 || y >= j1 + PONG_PADDLE_LENGTH)
			return -1;

		if(*ball_x > 2)
			move_ball(ball_x, ball_y, *ball_x - 2, 0);

		offset = y - (j1 + PONG_PADDLE_LENGTH / 2) + 1;
		if(abs(offset) < 2) {
			*dx = 2;
			*dy = 0;
		} else if(abs(offset) < 5) {
			*dx = 2;
			*dy = 1;
		} else if(abs(offset) < 8) {
			*dx = 1;
			*dy = 1;
		} else {
			*dx = 1;
			*dy = 2;
		}

		if(offset < 0)
			*dy = -(*dy);

		beep(1, 1);
	}

	//check computer's paddle
	if(x + PONG_BALL_SIZE > LCD_WIDTH - 2) {
		if(y + PONG_BALL_SIZE <= j2 - 1 || y >= j2 + PONG_PADDLE_LENGTH)
			return 1;

		if(*ball_x + PONG_BALL_SIZE < LCD_WIDTH - 2)
			move_ball(ball_x, ball_y, LCD_WIDTH - 2 - *ball_x - PONG_BALL_SIZE, 0);

		offset = y - (j2 + PONG_PADDLE_LENGTH / 2) + 1;
		if(abs(offset) < 2) {
			*dx = -2;
			*dy = 0;
		} else if(abs(offset) < 5) {
			*dx = -2;
			*dy = 1;
		} else if(abs(offset) < 8) {
			*dx = -1;
			*dy = 1;
		} else {
			*dx = -1;
			*dy = 2;
		}

		if(offset < 0)
			*dy = -(*dy);

		beep(1, 1);
	}

	//check top and bottom
	y = *ball_y + *dy;
	if(y < 0) {
		if(*ball_y > 0)
			move_ball(ball_x, ball_y, 0, -(*ball_y));
		*dy = -(*dy);
	}
	if(y + PONG_BALL_SIZE - 1 >= LCD_HEIGHT) {
		if(*ball_y + PONG_BALL_SIZE < LCD_HEIGHT)
			move_ball(ball_x, ball_y, 0, LCD_HEIGHT - (*ball_y) - PONG_BALL_SIZE);
		*dy = -(*dy);
	}
	return 0;
}

//does a missaligin so that th ball won't hit the center every time
void dirty_ball_move(uint8_t *ball_x, uint8_t *ball_y, int dx, int dy) {
	if(rand() % 4 == 0) {
		if(dy > 0 && *ball_y + PONG_BALL_SIZE < LCD_HEIGHT)
			move_ball(ball_x, ball_y, 0, 1);
	} else if(rand() % 2 == 0 && dy < 0 && *ball_y > 0)
		move_ball(ball_x, ball_y, 0, -1);
}

void display_score(uint8_t score1, uint8_t score2) {
	int i;

	clcd_init();
	cdisplay_clear();

	display_line(padding_string);
	display_line(padding_string);
	for(i = 0; i < 5; i++)
		write_data(' ');
	write_data('0' +  score1);
	write_data(' ');
	write_data('-');
	write_data(' ');
	write_data('0' +  score2);

	_delay_ms(1000);
	cdisplay_clear();
}

void pong() {
	uint8_t i1, j1, i2, j2;
	uint8_t ball_x, ball_y;
	uint8_t score1, score2;
	int dx, dy, winner;

	dx = -2;
	dy = 0;
	score1 = 0;
	score2 = 0;

	display_score(0, 0);

	while(!menu) {
		i1 = 0;
		j1 = (LCD_HEIGHT - PONG_PADDLE_LENGTH) / 2;
		i2 = LCD_WIDTH - 2;
		j2 = (LCD_HEIGHT - PONG_PADDLE_LENGTH) / 2;

		ball_x = 62;
		ball_y = 30;
		

		glcd_enable();
		glcd_fill(0x0000);

		draw_vertical_line(i1, j1, PONG_PADDLE_LENGTH);
		draw_vertical_line(i1 + 1, j1, PONG_PADDLE_LENGTH);

		draw_vertical_line(i2, j2, PONG_PADDLE_LENGTH);
		draw_vertical_line(i2 + 1, j2, PONG_PADDLE_LENGTH);

		draw_ball(ball_x, ball_y);

		while(!menu) {
			check_input();

			if(up)
				j1 = move_paddle_up(i1, j1);
			else if(down)
				j1 = move_paddle_down(i1, j1);

			dirty_ball_move(&ball_x, &ball_y, dx, dy);

			winner = detect_collision(&ball_x, &ball_y, &dx, &dy, j1, j2);
			if(winner != 0) {
				move_ball(&ball_x, &ball_y, dx, dy);
				
				if(winner == -1) {
					dx = 2;
					dy = 0;
					score2++;
				} else {
					dx = -2;
					dy = 0;
					score1++;
				}

				_delay_ms(20);
				glcd_fill(0x0000);
				display_score(score1, score2);

				break;
			}

			move_ball(&ball_x, &ball_y, dx, dy);

			if(rand() % 10 != 0) {
				if(ball_y < j2 + PONG_PADDLE_LENGTH / 2)
					j2 = move_paddle_up(i2, j2);
				else if(ball_y > j2 + PONG_PADDLE_LENGTH / 2)
					j2 = move_paddle_down(i2, j2);
			}
		}

		if(score1 == NO_PONG_GAMES_TO_WIN) {
			display_text(padding_string, you_won_string, padding_string, padding_string);
			win_sound();
			wait_for_input();
			cdisplay_clear();
			_delay_ms(200);
			return;
		}
		if(score2 == NO_PONG_GAMES_TO_WIN) {
			display_text(padding_string, you_lost_string, padding_string, padding_string);
			lose_sound();
			wait_for_input();
			cdisplay_clear();
			_delay_ms(200);
			return;
		}	
	}

	glcd_fill(0x0000);
}

int main(void) {

	int game;

	DATAPORT_DIR = 0xff;
	DATA_PORT = 0x00;
	COMMAND_DIR = 0x07;
	COMMAND_PORT = 0x00;
  	DDRD |= 1 << PD5;

  	while(1) {
  		game = game_menu();

  		if(game == SNAKE)
  			snake();
  		else if(game == PONG)
  			pong();

  		_delay_ms(500);
  	}


	return 0;
}
